import asyncio

from a_sync._helpers import get_event_loop


def test_get_event_loop():
    assert get_event_loop() == asyncio.get_event_loop()

def test_get_event_loop_in_thread():
    def task():
        assert get_event_loop() == asyncio.get_event_loop()
    loop = get_event_loop()
    loop.run_until_complete(loop.run_in_executor(None, task))



def work():
    import time
    time.sleep(5)

from a_sync import ProcessPoolExecutor

@pytest.marks.asyncio
async def test_executor():
    executor = ProcessPoolExecutor(6)
    await executor.run(work())
    await executor.submit(work())



import asyncio
import sys

from a_sync._typing import *

if sys.version_info < (3, 9):
    class _Queue(asyncio.Queue, Generic[T]):...
else:
    class _Queue(asyncio.Queue[T]):...

class Queue(_Queue[T]):
    """The only difference between an a_sync.Queue and an asyncio.Queue is that `get_nowait` can retrn multiple responses."""
    @overload
    async def get(self, i: int = 1, can_return_less: bool = False) -> T:
        ...
    @overload
    async def get(self, i: int, can_return_less: bool = False) -> List[T]:
        ...
    async def get(self, i: int = 1, can_return_less: bool = False) -> Union[T, List[T]]:
        _validate_args(i, can_return_less)
        if i == 1:
            return await super().get()
        try:
            items = self.get_nowait(i, can_return_less=True)
        except asyncio.QueueEmpty:
            items = [await super().get()]
        if len(items) == i or can_return_less:
            return items
        while len(items) < i:
            items.extend(await self.get(i - len(items)))
        return items
        
    @overload
    def get_nowait(self, i: int = 1, can_return_less: bool = False) -> T:
        ...
    @overload
    def get_nowait(self, i: int, can_return_less: bool = False) -> List[T]:
        ...
    def get_nowait(self, i: int = 1, can_return_less: bool = False) -> Union[T, List[T]]:
        """
        Just like `asyncio.Queue.get_nowait`, but will return `i` items instead of 1.
        Set `can_return_less` to True if you want to receive up to `i` items.
        """
        _validate_args(i, can_return_less)
        values = []
        if i == -1:
            while True:
                try:
                    values.append(super().get_nowait())
                except asyncio.QueueEmpty:
                    return values
        for _ in range(i):
            try:
                values.append(super().get_nowait())
            except asyncio.QueueEmpty:
                if can_return_less:
                    break
                for value in values:
                    self.put_nowait(value)
                raise
        return values[0] if i == 1 else values

def _validate_args(i: int, can_return_less: bool) -> None:
    if not isinstance(i, int):
        raise TypeError(f"`i` must be a non-zero integer. You passed {i}")
    if not isinstance(can_return_less, bool):
        raise TypeError(f"`can_return_less` must be boolean. You passed {can_return_less}")
    if i == 0:
        raise ValueError(f"`i` must be a non-zero integer. You passed {i}")
    if can_return_less and i == 1:
        raise ValueError("you can't set i == 1 with can_return_less == True")


# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import asyncio

from a_sync import exceptions
from a_sync._typing import *
from a_sync.modifiers.cache.memory import apply_async_memory_cache


class CacheArgs(TypedDict):
    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[int]

@overload
def apply_async_cache(
    coro_fn: Literal[None],
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_async_cache(
    coro_fn: int,
    **modifiers: Unpack[CacheArgs],
) -> AsyncDecorator[P, T]:...
    
@overload
def apply_async_cache(
    coro_fn: CoroFn[P, T],
    **modifiers: Unpack[CacheArgs],
) -> CoroFn[P, T]:...
    
def apply_async_cache(
    coro_fn: Union[CoroFn[P, T], CacheType, int] = None,
    cache_type: CacheType = 'memory',
    cache_typed: bool = False,
    ram_cache_maxsize: Optional[int] = None,
    ram_cache_ttl: Optional[int] = None,
) -> AsyncDecoratorOrCoroFn[P, T]:
    
    # Parse Inputs
    if isinstance(coro_fn, int):
        assert ram_cache_maxsize is None
        ram_cache_maxsize = coro_fn
        coro_fn = None
    
    # Validate 
    elif coro_fn is None:
        if ram_cache_maxsize is not None and not isinstance(ram_cache_maxsize, int):
            raise TypeError("'lru_cache_maxsize' must be an integer or None.", ram_cache_maxsize)
    elif not asyncio.iscoroutinefunction(coro_fn):
        raise exceptions.FunctionNotAsync(coro_fn)
    
    if cache_type == 'memory':
        cache_decorator = apply_async_memory_cache(maxsize=ram_cache_maxsize, ttl=ram_cache_ttl, typed=cache_typed)
        return cache_decorator if coro_fn is None else cache_decorator(coro_fn)
    elif cache_type == 'disk':
        pass
    raise NotImplementedError(f"cache_type: {cache_type}")



from typing import Any

from a_sync import exceptions

AFFIRMATIVE_FLAGS = {'sync'}
NEGATIVE_FLAGS = {'asynchronous'}
VIABLE_FLAGS = AFFIRMATIVE_FLAGS | NEGATIVE_FLAGS

"""
A-Sync uses 'flags' to indicate whether objects / fn calls will be sync or async.
You can use any of the provided flags, whichever makes most sense for your use case.
"""

def negate_if_necessary(flag: str, flag_value: bool):
    validate_flag_value(flag, flag_value)
    if flag in AFFIRMATIVE_FLAGS:
        return bool(flag_value)
    elif flag in NEGATIVE_FLAGS:
        return bool(not flag_value)
    raise exceptions.InvalidFlag(flag)

def validate_flag_value(flag: str, flag_value: Any) -> bool:
    if not isinstance(flag_value, bool):
        raise exceptions.InvalidFlagValue(flag, flag_value)
    return flag_value




import asyncio

import async_property as ap  # type: ignore [import]

from a_sync import config
from a_sync._typing import *
from a_sync.modified import Modified
from a_sync.modifiers.manager import ModifierManager


class PropertyDescriptor(Modified[T]):
    def __init__(self, _fget: Callable[..., T], field_name=None, **modifiers: ModifierKwargs):
        if not callable(_fget):
            raise ValueError(f'Unable to decorate {_fget}')
        self.modifiers = ModifierManager(modifiers)
        self._fn = _fget
        _fget = self.modifiers.apply_async_modifiers(_fget) if asyncio.iscoroutinefunction(_fget) else self._asyncify(_fget)
        super().__init__(_fget, field_name=field_name)  # type: ignore [call-arg]
    def __repr__(self) -> str:
        return f"<{self.__class__.__module__}.{self.__class__.__name__} for {self._fn} at {hex(id(self))}>"
    
class AsyncPropertyDescriptor(PropertyDescriptor[T], ap.base.AsyncPropertyDescriptor):
    pass
        
class AsyncCachedPropertyDescriptor(PropertyDescriptor[T], ap.cached.AsyncCachedPropertyDescriptor):
    pass


@overload
def a_sync_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> Callable[[Property[T]], AsyncPropertyDescriptor[T]]:...
    
@overload
def a_sync_property(  # type: ignore [misc]
    func: Property[T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> AsyncPropertyDescriptor[T]:...
    
def a_sync_property(  # type: ignore [misc]
    func: Union[Property[T], DefaultMode] = None,
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    AsyncPropertyDescriptor[T],
    Callable[[Property[T]], AsyncPropertyDescriptor[T]],
]:
    if func in ['sync', 'async']:
        modifiers['default'] = func
        func = None
    def modifier_wrap(func: Property[T]) -> AsyncPropertyDescriptor[T]:
        return AsyncPropertyDescriptor(func, **modifiers)
    return modifier_wrap if func is None else modifier_wrap(func)  # type: ignore [arg-type]
    

@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Literal[None],
    default: DefaultMode,
    **modifiers: Unpack[ModifierKwargs],
) -> Callable[[Property[T]], AsyncCachedPropertyDescriptor[T]]:...
    
@overload
def a_sync_cached_property(  # type: ignore [misc]
    func: Property[T],
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> AsyncCachedPropertyDescriptor[T]:...
    
def a_sync_cached_property(  # type: ignore [misc]
    func: Optional[Property[T]] = None,
    default: DefaultMode = config.DEFAULT_MODE,
    **modifiers: Unpack[ModifierKwargs],
) -> Union[
    AsyncCachedPropertyDescriptor[T],
    Callable[[Property[T]], AsyncCachedPropertyDescriptor[T]],
]:
    def modifier_wrap(func: Property[T]) -> AsyncCachedPropertyDescriptor[T]:
        return AsyncCachedPropertyDescriptor(func, **modifiers)
    return modifier_wrap if func is None else modifier_wrap(func)




from functools import cached_property
from logging import Logger, getLogger, DEBUG


class _LoggerMixin:
    @cached_property
    def logger(self) -> Logger:
        logger_id = f"{self.__class__.__module__}.{self.__class__.__name__}"
        if hasattr(self, '_name') and self._name:
            logger_id += f'.{self._name}'
        return getLogger(logger_id)
    @property
    def debug_logs_enabled(self) -> bool:
        return self.logger.isEnabledFor(DEBUG)



import asyncio

from a_sync.utils.iterators import (as_yielded, exhaust_iterator,
                                    exhaust_iterators)
from a_sync.utils.as_completed import as_completed
from a_sync.utils.gather import gather


__all__ = [
    "all",
    "any",
    "as_completed",
    "as_yielded",
    "gather",
    "exhaust_iterator",
    "exhaust_iterators",
]

async def any(*awaitables) -> bool:
    futs = [asyncio.ensure_future(a) for a in awaitables]
    for fut in asyncio.as_completed(futs):
        if bool(await fut):
            for fut in futs:
                fut.cancel()
            return True
    return False
    
async def all(*awaitables) -> bool:
    futs = [asyncio.ensure_future(a) for a in awaitables]
    for fut in asyncio.as_completed(awaitables):
        if not bool(await fut):
            for fut in futs:
                fut.cancel()
            return False
    return True



# mypy: disable-error-code=valid-type
# mypy: disable-error-code=misc
import functools
import sys

from a_sync import _helpers, _kwargs
from a_sync._typing import *
from a_sync.modifiers.manager import ModifierManager
    

class Modified(Generic[T]):
    modifiers: ModifierManager
    def _asyncify(self, func: SyncFn[P, T]) -> CoroFn[P, T]:
        """Applies only async modifiers."""
        coro_fn = _helpers._asyncify(func, self.modifiers.executor)
        return self.modifiers.apply_async_modifiers(coro_fn)
    @functools.cached_property
    def _await(self): # -> SyncFn[[Awaitable[T]]] T]:
        """Applies only sync modifiers."""
        return self.modifiers.apply_sync_modifiers(_helpers._await)
    @property
    def default(self) -> DefaultMode:
        return self.modifiers.default
    
        
class ASyncFunction(Modified[T], Callable[P, T], Generic[P, T]):
    @overload
    def __init__(self, fn: CoroFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:...
    @overload
    def __init__(self, fn: SyncFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:...
    def __init__(self, fn: AnyFn[P, T], **modifiers: Unpack[ModifierKwargs]) -> None:
        _helpers._validate_wrapped_fn(fn)
        self.modifiers = ModifierManager(modifiers)
        self.__wrapped__ = fn
        if hasattr(self.__wrapped__, '__name__'):
            self.__name__ = self.__wrapped__.__name__
    
    @overload
    def __call__(self, *args: P.args, sync: Literal[True] = True, **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, sync: Literal[False] = False, **kwargs: P.kwargs) -> Awaitable[T]:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[False] = False, **kwargs: P.kwargs) -> T:...
    @overload
    def __call__(self, *args: P.args, asynchronous: Literal[True] = True, **kwargs: P.kwargs) -> Awaitable[T]:...
    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:
        return self.fn(*args, **kwargs)
    
    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.__name__} at {hex(id(self))}>"
    
    @functools.cached_property
    def fn(self): # -> Union[SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]], SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]]:
        """Returns the final wrapped version of 'self._fn' decorated with all of the a_sync goodness."""
        return self._async_wrap if self._async_def else self._sync_wrap
    
    @property
    def _sync_default(self) -> bool:
        """If user did not specify a default, we defer to the function. 'def' vs 'async def'"""
        return True if self.default == 'sync' else False if self.default == 'async' else not self._async_def
    
    @property
    def _async_def(self) -> bool:
        return asyncio.iscoroutinefunction(self.__wrapped__)
    
    def _run_sync(self, kwargs: dict):
        if flag := _kwargs.get_flag_name(kwargs):
            # If a flag was specified in the kwargs, we will defer to it.
            return _kwargs.is_sync(flag, kwargs, pop_flag=True)
        else:
            # No flag specified in the kwargs, we will defer to 'default'.
            return self._sync_default
    
    @functools.cached_property
    def _asyncified(self) -> CoroFn[P, T]:
        """Turns 'self._fn' async and applies both sync and async modifiers."""
        assert not self._async_def, f"Can only be applied to sync functions, not {self.__wrapped__}"
        return self._asyncify(self._modified_fn)  # type: ignore [arg-type]
    
    @functools.cached_property
    def _modified_fn(self) -> AnyFn[P, T]:
        """
        Applies sync modifiers to 'self._fn' if 'self._fn' is a sync function.
        Applies async modifiers to 'self._fn' if 'self._fn' is a sync function.
        """
        if self._async_def:
            return self.modifiers.apply_async_modifiers(self.__wrapped__)  # type: ignore [arg-type]
        return self.modifiers.apply_sync_modifiers(self.__wrapped__)  # type: ignore [return-value]
    
    @functools.cached_property
    def _async_wrap(self): # -> SyncFn[[CoroFn[P, T]], MaybeAwaitable[T]]:
        """The final wrapper if self._fn is an async function."""
        @functools.wraps(self._modified_fn)
        def async_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
            should_await = self._run_sync(kwargs) # Must take place before coro is created, we're popping a kwarg.
            coro = self._modified_fn(*args, **kwargs)
            return self._await(coro) if should_await else coro
        return async_wrap
    
    @functools.cached_property
    def _sync_wrap(self): # -> SyncFn[[SyncFn[P, T]], MaybeAwaitable[T]]:
        """The final wrapper if self._fn is a sync function."""
        @functools.wraps(self._modified_fn)
        def sync_wrap(*args: P.args, **kwargs: P.kwargs) -> MaybeAwaitable[T]:  # type: ignore [name-defined]
            if self._run_sync(kwargs):
                return self._modified_fn(*args, **kwargs)
            return self._asyncified(*args, **kwargs)            
        return sync_wrap
    
if sys.version_info < (3, 10):
    _inherit = ASyncFunction[AnyFn[P, T], ASyncFunction[P, T]]
else:
    _inherit = ASyncFunction[[AnyFn[P, T]], ASyncFunction[P, T]]
              
class ASyncDecorator(_inherit[P, T]):
    _fn = None
    def __init__(self, **modifiers: Unpack[ModifierKwargs]) -> None:
        assert 'default' in modifiers, modifiers
        self.modifiers = ModifierManager(modifiers)
        self.validate_inputs()
        
    def validate_inputs(self) -> None:
        if self.modifiers.default not in ['sync', 'async', None]:
            raise ValueError(f"'default' must be either 'sync', 'async', or None. You passed {self.modifiers.default}.")
        
    def __call__(self, func: AnyFn[P, T]) -> ASyncFunction[P, T]:  # type: ignore [override]
        return ASyncFunction(func, **self.modifiers)
